"""
delimited.path
~~~~~~~~~~~~~~

This module defines Path objects. a Path object implements an interface though
which a path to nested data can be defined using one or more path segments.
"""

import abc
import copy

from delimited.sequence import ListIndex


class Path(abc.ABC):
    """ The abstract base class for Path objects. When subclassing Path the
    _encode and _decode methods must be overridden to handle a collapsed path
    format. Path segments are stored internally in a list in the
    segments attribute.
    """

    def __init__(self, path=None):
        self(path=path)

    def __call__(self, path=None):
        """ Overwrite path segments with decoded path param. If path param is
        None, set path segments to an empty list.
        """

        if path is None:
            self.segments = []

        else:
            if not isinstance(path, list):
                path = self._decode(path)
            self.segments = path

    def __add__(self, other):
        """ Add other to the end of path segments. Accepts a path segment.
        """

        self.segments = self.segments + [other]
        return self

    def __iadd__(self, other):
        """ Add other to the end of path segments. Accepts a path segment.
        """

        return self.__add__(other)

    def __eq__(self, other):
        """ Compare equality with instance of self or path in collapsed path
        format.
        """

        if isinstance(other, self.__class__):
            return self.segments == other.segments
        return self.encode() == other

    def __ne__(self, other):
        """ Compare inequality with instance of self or path in collapsed path
        format.
        """

        return not self.__eq__(other)

    def __hash__(self):
        """ Return hash of self.encode().
        """

        return hash(self.encode())

    def __len__(self):
        """ Return length of path segments.
        """

        return len(self.segments)

    def __bool__(self):
        """ Return length of path segments cast to boolean.
        """

        return bool(len(self.segments))

    def __str__(self):
        """ Return self.encode() cast to str.
        """

        return str(self.encode())

    def __repr__(self):
        """ Return class name joined with encoded path segments cast to str.
        """

        return f"{self.__class__.__name__}({self.encode()})"

    def __iter__(self):
        """ Yield path segments.
        """

        for k in self.segments:
            yield k

    def __reversed__(self):
        """ Yield path segments in reverse order.
        """

        for k in self.segments[::-1]:
            yield k

    def __contains__(self, key):
        """ Return boolean for presence of key in internal path segments.
        Accepts path segment.
        """

        return key in self.segments

    def __getitem__(self, i):
        """ Return item at index or slice in collapsed path format.
        """

        value = self.segments[i]
        if isinstance(value, list):
            value = self._encode(value)
        return value

    def __setitem__(self, i, value):
        """ Set item at index to value.
        """
        self.segments[i] = value

    def __delitem__(self, i):
        """ Delete item at index.
        """

        del self.segments[i]

    @classmethod
    @abc.abstractmethod
    def _encode(self, value):
        """ Encode list to collapsed path. This method must be overridden
        when subclassing.
        """

        pass  # pragma: no cover

    @classmethod
    @abc.abstractmethod
    def _decode(self, value):
        """ Decode a value in collapsed path format to a list. This method
        must be overridden when subclassing.
        """

        pass  # pragma: no cover

    @property
    def head(self):
        """ Return the head of the list of path segments, meaning all segments
        except for the last one. If there is only one segment return None
        """

        h = self.segments[:-1]
        return self._encode(h) if isinstance(h, list) and h else h or None
    
    @property
    def tail(self):

        return self.segments[-1]

    def append(self, value):
        """ Add value to the end of path segments. Accepts a path segment.
        """

        return self.segments.append(value)

    def extend(self, values):
        """ Add each item of values to the end of path segments. Accepts an
        instance of self or a encoded group of path segments.
        """

        if isinstance(values, self.__class__):
            values = values.segments
        else:
            values = self._decode(values)
        return self.segments.extend(values)

    def insert(self, i, value):
        """ Insert value at index i in path segments.
        """

        self.segments.insert(i, value)

    def remove(self, value):
        """ Remove the first item from path segments that is equal to value.
        Raise an exception if value is not found.
        """

        return self.segments.remove(value)

    def pop(self, *args):
        """ Remove the item at index i from path segments and return. If i is
        not given, remove and return the first value in self.segments.
        """

        return self.segments.pop(*args)

    def clear(self):
        """ Remove all values from path segments.
        """

        self()

    def index(self, value):
        """ Return the index from path segments of the first item that is
        equal to value. Raise an exception if value is not found.
        """

        return self.segments.index(value)

    def count(self, value):
        """ Return the number of times value appears in path segments.
        """

        return self.segments.count(value)

    def reverse(self):
        """ Reverse path segments in place.
        """

        self.segments = self.segments[::-1]

    def copy(self, i=None):
        """ Return an instance of self.__class__ with its path segments
        set to a shallow copy of this instances path segments resolved
        index using i if passed.
        """

        segments = copy.copy(self.segments if i is None else self.segments[i])
        return self.__class__(segments)

    def clone(self):
        """ Return an instance of self.__class__ with its path segments
        set to a reference of this instances path segments.
        """

        return self.__class__(self.segments)

    def encode(self):
        """ Call self._encode with path segments and return result.
        """

        return self._encode(self.segments)


class TuplePath(Path):
    """ This class implements tuple path notation as its collapsed path format.
    TuplePaths can handle any hashable type as a path segment.
    Example: ("key1", "key2", "key3")
    """

    @classmethod
    def _encode(self, value):
        """ Encode a list to collapsed path format.
        """

        return tuple(value)

    @classmethod
    def _decode(self, value):
        """ Decode collapsed path format to a list.
        """

        return list(value) if isinstance(value, tuple) else [value]


class DelimitedStrPath(Path):
    """ This class implements delimited string path notation as its collapsed
    path format.
    Example: "key1.key2.key3"
    """

    delimiter = "."

    @classmethod
    def _encode(self, value):
        """ Encode a list to collapsed path format.
        """
        
        if not isinstance(value, list):
            value = [value]
            
        for i, v in enumerate(value):
            if isinstance(v, ListIndex):
                value[i] = f"::{v.index}"

        return self.delimiter.join(value)

    @classmethod
    def _decode(self, value):
        """ Decode collapsed path format to a list.
        """
        
        decoded = []
        for v in value.split(self.delimiter):
            if v[:2] == "::":
                v = ListIndex(int(v[2:]))
            decoded.append(v)
            
        return decoded

    def __repr__(self):
        """ Custom repr for string based path.
        """

        return f"{self.__class__.__name__}('{self.encode()}')"



"""
delimited.container
~~~~~~~~~~~~~~~~~
"""

import copy

from delimited.sequence import SequenceIndex
from delimited.sequence import SequenceValue


class NestedContainer(object):
    container = None
    path = None
    sequence = None
    mapping = None
    
    def __init__(self, data=None):
        self(data)
        
    def __call__(self, data=None):
        if data is None:
            self.data = self.container()
        elif isinstance(data, self.container):
            self.data = self._wrap(data)
        else:
            raise TypeError(f"data is not of type {self.container}")
    
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.data == other.data
        return self.data == other
        
    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return str(self.data)
        
    def __repr__(self):
        return f"{self.__class__.__name__}({self.data})"
    
    def __hash__(self):
        return hash(str(self.data))

    def __len__(self):
        return len(self.data)

    def __bool__(self):
        return bool(len(self.data))
    
    def __contains__(self, value):
        return value in self.data
    
    def __getitem__(self, path):
        return self.ref(path)

    def __setitem__(self, path, value):
        self.set(path, value)

    def __delitem__(self, path):
        self.unset(path)

    def __copy__(self):
        new = self.__class__()
        new.data = copy.copy(self.data)
        return new

    def __deepcopy__(self, *args):
        new = self.__class__()
        new.data = copy.deepcopy(self.data)
        return new

    def _wrap(self, data):
        i = data.items() if isinstance(data, dict) else enumerate(data)
        for key, value in i:    
            if isinstance(value, dict):
                data[key] = self.mapping(data[key])
            elif isinstance(value, list):
                data[key] = self.sequence(data[key])
                
        return data
        
    def _unwrap(self, data):
        i = data.items() if isinstance(data, dict) else enumerate(data)
        for key, value in i:
            if isinstance(value, (self.mapping, self.sequence)):
                data[key] = value.get()
    
        return data
        
    def unwrap(self):
        return self._unwrap(self.data)

    def ref(self, path=None, create=False):
        if path is None:
            return self.data

        haystack = self.data

        if not isinstance(path, self.path):
            path = self.path(path)

        for i, segment in enumerate(path):
            
            s = segment.index if isinstance(segment, SequenceIndex) else segment
            
            try:
                haystack = haystack[s]

            except KeyError as e:
                if create:
                    haystack[s] = self.container()
                    haystack = haystack[s]

                else:
                    e.args = (f"{s} in {path}",) + e.args[1:]
                    raise
            
            except IndexError as e:
                if create:
                    haystack.append(self.container())
                    haystack = haystack[s]

                else:
                    e.args = (f"{s} in {path}",) + e.args[1:]
                    raise

            except TypeError as e:
                e.args = (f"{s} in {path}",) + e.args[1:]
                raise

        return haystack

    def get(self, path=None, *args):
        try:
            return copy.deepcopy(self._unwrap(self.ref(path)))
        except (KeyError, IndexError):
            if args:
                return args[0]
            raise

    def has(self, path=None):
        try:
            return bool(self.ref(path))
        except (KeyError, IndexError):
            return False

    def copy(self, path=None):
        return copy.copy(self)

    def clone(self, path=None):
        clone = self.__class__()
        clone.data = self.ref(path)
        return clone
    
    @classmethod
    def _merge(cls, a, b):
        if not isinstance(a, b.__class__):
            return copy.copy(a)
        
        b = copy.copy(b)
        iterable = a.keys() if isinstance(a, dict) else range(len(a))
        
        for k in iterable:
            
            # sequences
            if all(isinstance(v, list) for v in [a, b]):
                
                # unequal lengths
                if k > (len(b) - 1):
                    b.append(a[k])
                
                # recursive merge
                else:
                    b[k] = cls._merge(a[k], b[k])
                    
            # mappings
            elif all(isinstance(v, dict) for v in [a, b]):
                
                # key not set
                if k not in b:
                    b[k] = a[k]
                
                # recursive merge
                else:
                    b[k] = cls._merge(a[k], b[k])
        
        return b
    
    def merge(self, data, path=None):
        if isinstance(data, self.__class__):
            data = data.unwrap()
        
        return self._merge(self.get(path), data)
    
    def expand(self, data):
        self.data = self._expand(data)
        return self
    
    def _expand(self, data):
    
        # determine type for expanded data
        expanded = self.container()
    
        for path, value in data.items():
            if not isinstance(path, self.path):
                path = self.path(path)
    
            for i, segment in enumerate(reversed(path)):
    
                if isinstance(segment, SequenceIndex):
                    index = segment.index
                    new_segment = [SequenceValue()] * (index + 1)
    
                else:
                    index = segment
                    new_segment = dict()
    
                # first segment
                if i == 0:
                    new_segment[index] = value
                    expanded_segment = new_segment
    
                # > first segment
                elif i > 0:
                    new_segment[index] = expanded_segment
                    expanded_segment = new_segment
    
                # last segment
                if i == (len(path) - 1):
                    expanded = self._merge(expanded, expanded_segment)
    
        return expanded

    def collapse(self, path=None, func=None):
        data = self if path is None else self.ref(path)
        return self._collapse(data, func=func)
    
    @classmethod
    def _collapse(cls, data, func=None, _parent_path=None):
        """ Recursively collapse expanded nested data and return.
        """

        collapsed = {}
        path = cls.path() if _parent_path is None else _parent_path

        # sequence
        if isinstance(data, cls.sequence):
            i, index = enumerate(data), cls.sequence.sequenceindex

        # mapping
        elif isinstance(data, cls.mapping):
            i, index = data.items(), None

        # all others
        else:
            return data

        for key, value in i:
            
            # NOTE: use of `func` results in current level not being collapsed
            if callable(func) and func(key, value, data.__class__):

                if isinstance(data, cls.sequence):
                    uncollapsed = list([SequenceValue()] * (key + 1))
                    
                elif isinstance(data, cls.mapping):
                    uncollapsed = dict()

                # assign collapsed value to container
                # NOTE: by omitting `_parent_path` in collapse call below and 
                # assiging directly to container, all paths of collapsed
                # children will not include current level path
                uncollapsed[key] = cls._collapse(value, func=func)
                value = {path.encode(): uncollapsed} if len(path) else uncollapsed

            else:

                # wrap key in sequence index class if necessary
                if index is not None:
                    key = index(key)

                # NOTE: create copy of path so further extend calls do not
                # mutate current level path
                path_copy = path.copy()
                path_copy.extend(key)

                # continue to collapse nested values
                value = cls._collapse(value, func=func, _parent_path=path_copy)

                # if value is not collapsable, tree has been traversed for branch
                # and value of `encoded path: value` should be returned
                if not isinstance(value, (list, dict)) and len(path_copy):
                    value = {path_copy.encode(): value}

            # update root with branch
            collapsed = cls._merge(collapsed, value)
            # collapsed.update(value)

        return collapsed
    
    def set(self, path, value, create=True):
        if not isinstance(path, self.path):
            path = self.path(path)
            
        tail = path.tail
        if path.head is None and isinstance(path.tail, SequenceIndex):
            tail = path.tail.index
            
        haystack = self.ref(path.head, create=create)
        haystack[tail] = value

    def unset(self, path, cleanup=False):
        if not isinstance(path, self.path):
            path = self.path(path)
        
        haystack = self.ref(path.head)
        
        tail = path.tail
        if path.head is None and isinstance(path.tail, SequenceIndex):
            tail = path.tail.index
        
        del haystack[tail]

        if cleanup and path.head is not None:
            if not len(haystack):
                self.unset(path.head)

        return True

    def push(self, path, value, create=True):
        if not isinstance(path, self.path):
            path = self.path(path)

        haystack = self.ref(path.head, create=create)
        
        tail = path.tail
        if path.head is None and isinstance(path.tail, SequenceIndex):
            tail = path.tail.index

        try:
            haystack[tail].append(value)

        except KeyError as e:
            if create:
                haystack[tail] = self.sequence()
                haystack[tail].append(value)
            else:
                e.args = (f"{tail} in {path}",) + e.args[1:]
                raise

        except AttributeError as e:
            if create:
                haystack[tail] = self.sequence([haystack[tail]])
                haystack[tail].append(value)
            else:
                e.args = (f"{tail} in {path}",) + e.args[1:]
                raise

        return True

    def pull(self, path, value, cleanup=False):
        if not isinstance(path, self.path):
            path = self.path(path)

        haystack = self.ref(path.head)

        tail = path.tail
        if path.head is None and isinstance(path.tail, SequenceIndex):
            tail = path.tail.index

        haystack[tail].remove(value)

        if cleanup:
            if not len(haystack[tail]):
                del haystack[tail]

        return True




import sys; sys.path.append("../")

import unittest
import copy

from delimited.path import TuplePath
from delimited.path import DelimitedStrPath

from delimited.sequence import ListIndex
from delimited.sequence import SequenceValue

from delimited import NestedList


class TestListIndex(unittest.TestCase):

    # __init__

    def test____init__(self):
        a = ListIndex(0)
        self.assertEqual(a.index, 0)
        
    def test____init____raises_TypeError(self):
        with self.assertRaises(TypeError):
            a = ListIndex("foo")

    # __str__
    
    def test____str__(self):
        a = ListIndex(0)
        self.assertEqual(str(a), "0")
        
    # __repr__
    
    def test____repr__(self):
        a = ListIndex(0)
        self.assertEqual(repr(a), "ListIndex(0)")
        
    # __eq__
    
    def test___eq____same_type(self):
        self.assertEqual(ListIndex(0), ListIndex(0))
        
    def test___eq____different_type(self):
        self.assertEqual(ListIndex(0), 0)
        
    # __ne__
    
    def test___ne____same_type(self):
        self.assertNotEqual(ListIndex(0), ListIndex(1))
        
    def test___ne____different_type(self):
        self.assertNotEqual(ListIndex(0), 1)
        self.assertNotEqual(ListIndex(0), "foo")
        
    # __hash__
    
    def test___hash__(self):
        a1 = hash(ListIndex(0))
        a2 = hash(ListIndex(0))
        self.assertTrue(a1 == a2)


class TestNestedList(unittest.TestCase):

    # __init__

    def test___init____no_args(self):
        a = NestedList()
        self.assertEqual(a.data, [])
    
    def test___init____list_arg(self):
        a = NestedList(["v"])
        self.assertEqual(a, ["v"])
    
    def test___init____raises_TypeError(self):
        with self.assertRaises(TypeError):
            a = NestedList("foo")
    
    # __call__
    
    def test___call____no_args(self):
        a = NestedList(["v"])
        a()
        self.assertEqual(a, [])
    
    def test___call____dict_arg(self):
        a = NestedList()
        a(["v"])
        self.assertEqual(a, ["v"])
    
    # __bool__
    
    def test___bool____empty_data__returns_False(self):
        self.assertFalse(bool(NestedList()))
    
    def test___bool____non_empty_data__returns_True(self):
        self.assertTrue(bool(NestedList(["v"])))
    
    # __repr__
    
    def test___repr__(self):
        a = NestedList(["v"])
        self.assertEqual(repr(a), "NestedList(['v'])")
    
    # __eq__
    
    def test___eq____returns_True(self):
        a1 = NestedList(["v"])
        a2 = NestedList(["v"])
        self.assertTrue(a1 == a2)
    
    def test___eq____same_class__returns_False(self):
        a1 = NestedList(["foo"])
        a2 = NestedList(["bar"])
        self.assertFalse(a1 == a2)
    
    def test___eq____different_class__returns_True(self):
        a1 = NestedList(["v"])
        a2 = ["v"]
        self.assertTrue(a1 == a2)
    
        a3 = NestedList()
        a4 = []
        self.assertTrue(a3 == a4)
    
    # __ne__
    
    def test___ne____returns_True(self):
        a1 = NestedList(["foo"])
        a2 = NestedList(["bar"])
        self.assertTrue(a1 != a2)
    
    def test___ne____returns_False(self):
        a1 = NestedList(["v"])
        a2 = NestedList(["v"])
        self.assertFalse(a1 != a2)
    
    # __str__
    
    def test___str__(self):
        a = NestedList(["v"])
        self.assertEqual(str(a), "['v']")
    
    # __hash__
    
    def test___hash__(self):
        a1 = hash(NestedList(["v"]))
        a2 = hash(NestedList(["v"]))
        self.assertTrue(a1 == a2)
    
    # __iter__
    
    def test___iter__(self):
        a = NestedList(["v"])
        for v in a:
            self.assertEqual(v, "v")
    
    # __contains__
    
    def test___contains__(self):
        a = NestedList(["v"])
        self.assertTrue("v" in a)
    
    # __getitem__
    
    def test___getitem__int_arg__returns_value(self):
        a = NestedList(["v"])
        self.assertEqual(a[ListIndex(0)], "v")
    
    def test___getitem__delimited_tuple_arg__returns_value(self):
        a = NestedList([[["v"]]])
        self.assertEqual(a[(ListIndex(0), ListIndex(0), ListIndex(0))], "v")
    
    # __setitem__
    
    def test___setitem__string_arg__sets_value(self):
        a = NestedList(["v"])
        a[ListIndex(0)] = "foo"
        self.assertEqual(a, ["foo"])
    
    def test___setitem__delimited_string_arg__sets_value(self):
        a = NestedList([[["v"]]])
        a[(ListIndex(0), ListIndex(0), ListIndex(0))] = "foo"
        self.assertEqual(a.data, [[["foo"]]])
    
    # __delitem__
    
    def test___delitem__string_arg__deletes_value(self):
        a = NestedList(["v"])
        del a[ListIndex(0)]
        self.assertEqual(a.data, [])
    
    def test___delitem__delimited_string_arg__deletes_value(self):
        a = NestedList([[["v"]]])
        del a[(ListIndex(0), ListIndex(0), ListIndex(0))]
        self.assertEqual(a.data, [[[]]])
    
    # __len__
    
    def test___len__(self):
        a = NestedList(["v"])
        self.assertEqual(len(a), 1)
    
    # __copy__
    
    def test___copy__(self):
        a1 = NestedList([[["foo"]]])
        a2 = copy.copy(a1)
        self.assertIsNot(a1, a2)
        self.assertEqual(a1, a2)
        a1.set((ListIndex(0), ListIndex(0), ListIndex(0)), "bar")
        self.assertEqual(a1, a2)
    
    # __deepcopy__
    
    def test___deepcopy__(self):
        a1 = NestedList([[["foo"]]])
        a2 = copy.deepcopy(a1)
        self.assertIsNot(a1, a2)
        self.assertEqual(a1, a2)
        a1.data[0][0][0] = "bar"
        self.assertNotEqual(a1, a2)
    
    # __add__
    
    def test___add____list_arg(self):
        a = NestedList(["v"])
        b = a + ["v"]
        self.assertEqual(b, ["v", "v"])
        
    def test___add____NestedList_arg(self):
        a = NestedList(["v"])
        b = a + NestedList(["v"])
        self.assertEqual(b, ["v", "v"])
        
    # __iadd__
    
    def test___iadd____list_arg(self):
        a = NestedList(["v"])
        a += ["v"]
        self.assertEqual(a, ["v", "v"])
        
    def test___iadd____NestedList_arg(self):
        a = NestedList(["v"])
        a += NestedList(["v"])
        self.assertEqual(a, ["v", "v"])
        
    # __iter__
    
    def test___iter__(self):
        a = NestedList(["v", "v", "v"])
        for v in a:
            self.assertEqual(v, "v")
        
    # __reversed__
    
    def test___reversed__(self):
        a = NestedList([3, 2, 1])
        for i, v in enumerate(reversed(a), 1):
            self.assertEqual(i, v)
        
    # append
    
    def test_append(self):
        a = NestedList(["v"])
        a.append("v")
        self.assertEqual(a, ["v", "v"])
        
    # extend
    
    def test_extend__list_arg(self):
        a = NestedList(["v"])
        a.extend(["v"])
        self.assertEqual(a, ["v", "v"])
        
    def test_extend__NestedList_arg(self):
        a = NestedList(["v"])
        a.extend(NestedList(["v"]))
        self.assertEqual(a, ["v", "v"])
    
    # insert
    
    def test_insert(self):
        a = NestedList(["bar"])
        a.insert(0, "foo")
        self.assertEqual(a, ["foo", "bar"])
    
    # remove
    
    def test_remove(self):
        a = NestedList(["v"])
        a.remove("v")
        self.assertEqual(a, [])
        
    # pop
    
    def test_pop(self):
        a = NestedList(["v"])
        self.assertEqual(a.pop(), "v")
        self.assertEqual(a, [])
        
    # clear
    
    def test_clear(self):
        a = NestedList(["v"])
        a.clear()
        self.assertEqual(a, [])
        
    # index
    
    def test_index(self):
        a = NestedList(["v"])
        self.assertEqual(a.index("v"), 0)
        
    # count
    
    def test_count(self):
        a = NestedList(["foo", "foo", "bar", "baz"])
        self.assertEqual(a.count("foo"), 2)
        
    # reverse
    
    def test_reverse(self):
        a = NestedList([3, 2, 1])
        a.reverse()
        self.assertEqual(a, [1, 2, 3])
    
    # ref
    
    def test_ref__no_arg__returns_all_attributes(self):
        a = NestedList(["v"])
        self.assertEqual(a.ref(), ["v"])
    
    def test_ref__string_arg__returns_value(self):
        a = NestedList(["v"])
        self.assertEqual(a.ref(ListIndex(0)), "v")
    
    def test_ref___raises_IndexError(self):
        a = NestedList()
        with self.assertRaises(IndexError):
            a.ref(ListIndex(0))
    
    def test_ref__long_path_arg__raises_ValueError(self):
        a = NestedList([["v"]])
        with self.assertRaises(IndexError):
            a.ref((ListIndex(0), ListIndex(1)))
    
    def test_ref__delimited_arg__returns_value(self):
        a = NestedList([[["v"]]])
        self.assertEqual(a.ref((ListIndex(0), ListIndex(0), ListIndex(0))), "v")
    
    def test_ref__True_create_arg_key_missing__creates_missing_containers(self):
        a = NestedList()
        a.ref((ListIndex(0), ListIndex(0)), create=True)
        self.assertEqual(a.data, [[[]]])
    
    # get
    
    def test_get__string_arg__returns_value(self):
        a = NestedList(["v"])
        self.assertEqual(a.get(0), "v")
    
    def test_get__string_arg_missing_key__returns_default_value(self):
        a = NestedList()
        self.assertEqual(a.get(0, "foo"), "foo")
    
    def test_get__string_arg__raises_IndexError(self):
        a = NestedList()
        with self.assertRaises(IndexError):
            a.get(0)
    
    def test_get__string_arg__raises_IndexError(self):
        a = NestedList([["v"]])
        with self.assertRaises(IndexError):
            a.get((ListIndex(0), ListIndex(0), ListIndex(1)))
    
    def test_get__path_arg__returns_value(self):
        a = NestedList([[["v"]]])
        self.assertEqual(a.get((ListIndex(0), ListIndex(0), ListIndex(0))), "v")
    
    # has
    
    def test_has__returns_True(self):
        self.assertTrue(NestedList(["v"]).has())
    
    def test_has__returns_False(self):
        self.assertFalse(NestedList().has())
    
    def test_has__path_arg__returns_True(self):
        a = NestedList([[["v"]]])
        self.assertTrue(a.has((ListIndex(0), ListIndex(0), ListIndex(0))))
    
    def test_has__path_arg__returns_False(self):
        a = NestedList([[["v"]]])
        self.assertFalse(a.has((ListIndex(1),)))
    
    # copy
    
    def test_copy(self):
        a1 = NestedList([[["v"]]])
        a2 = a1.copy()
        self.assertEqual(a1, a2)
        a1[ListIndex(0)] = "foo"
        self.assertNotEqual(a1, a2)
    
    # clone
    
    def test_clone(self):
        a1 = NestedList([[["v"]]])
        a2 = a1.clone()
        self.assertEqual(a1, a2)
        a1[ListIndex(0)] = "foo"
        self.assertEqual(a1, a2)
    
    # set
    
    def test_set__index_arg(self):
        a = NestedList(["foo"])
        a.set(ListIndex(0), "bar")
        self.assertEqual(a, ["bar"])
    
    def test_set__path_arg(self):
        a = NestedList([[["foo"]]])
        a.set((ListIndex(0), ListIndex(0), ListIndex(0)), "bar")
        self.assertEqual(a, [[["bar"]]])
    
    def test_set__raises_TypeError(self):
        a = NestedList([["foo"]])
        with self.assertRaises(TypeError):
            a.set((ListIndex(0), ListIndex(0), ListIndex(0)), "bar", create=False)
    
    # push
    
    def test_push__index_arg(self):
        a = NestedList([["foo"]])
        a.push(ListIndex(0), "bar")
        self.assertEqual(a, [["foo", "bar"]])
    
    def test_push__index_arg__convert_existing_value(self):
        a = NestedList(["foo"])
        a.push(ListIndex(0), "bar")
        self.assertEqual(a, [["foo", "bar"]])
    
    def test_push__path_arg(self):
        a = NestedList([[["foo"]]])
        a.push((ListIndex(0), ListIndex(0)), "bar")
        self.assertEqual(a, [[["foo", "bar"]]])
    
    def test_push__True_create_arg__creates_list(self):
        a = NestedList([[]])
        a.push(ListIndex(0), "bar")
        self.assertEqual(a, [["bar"]])
    
    def test_push__False_create_arg__raises_IndexError(self):
        a = NestedList()
        with self.assertRaises(IndexError):
            a.push(ListIndex(0), "foo", create=False)
    
    def test_push__False_create_arg__raises_TypeError(self):
        a = NestedList(["foo"])
        with self.assertRaises(AttributeError):
            a.push(ListIndex(0), "bar", create=False)
    
    # pull
    
    def test_pull__index_arg(self):
        a = NestedList([["foo"]])
        a.pull(ListIndex(0), "foo")
        self.assertEqual(a, [[]])
    
    def test_pull__cleanup_arg_True__removes_empty_containers(self):
        a = NestedList([[["v"]]])
        a.pull((ListIndex(0), ListIndex(0)), "v", cleanup=True)
        self.assertEqual(a, [[]])
    
    def test_pull__path_arg___raises_IndexError(self):
        a = NestedList()
        with self.assertRaises(IndexError):
            a.pull(ListIndex(0), "v")
    
    def test_pull__path_arg__raises_ValueError(self):
        a = NestedList([["foo"]])
        with self.assertRaises(ValueError):
            a.pull(ListIndex(0), "bar")
    
    def test_pull__path_arg__raises_TypeError(self):
        a = NestedList(["v"])
        with self.assertRaises(TypeError):
            a.pull((ListIndex(0), ListIndex(0)), "v")
    
    # unset
    
    def test_unset__path_arg(self):
        a = NestedList([[["v"]]])
        a.unset((ListIndex(0), ListIndex(0)))
        self.assertEqual(a, [[]])
    
    def test_unset__cleanup_arg_True__removes_empty_containers(self):
        a = NestedList([[["v"]]])
        a.unset((ListIndex(0), ListIndex(0)), cleanup=True)
        self.assertEqual(a, [])
    
    def test_unset__path_arg__raises_IndexError(self):
        a = NestedList([[["v"]]])
        with self.assertRaises(IndexError):
            a.unset((ListIndex(1), ListIndex(0)))
    
    # merge
    
    def test__merge(self):
        a = NestedList(["v"])
        self.assertEqual(["foo"], a.merge(["foo"]))
        
    def test__merge__unequal_length_sequences(self):
        a = NestedList(["foo", "bar"])
        self.assertEqual(["baz", "bar"], a.merge(["baz"]))

    def test_merge__nested_list_arg(self):
        a = NestedList([[["foo"]]])
        b = a.merge([[["bar"]]])
        self.assertEqual(b, [[["bar"]]])
    
    def test_merge__NestedDict_param(self):
        a = NestedList([[["foo"]]])
        b = a.merge(NestedList([[["bar"]]]))
        self.assertEqual(b, [[["bar"]]])
    
    def test_merge__incompatible_types(self):
        a = NestedList([[["v"]]])
        b = a.merge("foo")
        self.assertEqual(b, [[["v"]]])

    # collapse
    
    def test_collapse(self):
        a = NestedList(["v"])
        b = a.collapse()
        self.assertEqual(b, {(ListIndex(0),): 'v'})
        
    def test_collapse__function_arg(self):
        a = NestedList([[["v"]]])
    
        def detect_operator(key, value, container):
            return container is NestedList and value == "v"
        
        b = a.collapse(func=detect_operator)
        
        self.assertEqual(b, {(ListIndex(0), ListIndex(0)): ["v"]})

    # expand

    def test_expand(self):
        a = NestedList([[["v"]]])
        b = NestedList().expand(a.collapse())
        self.assertEqual(b, a)
        
    def test__expand(self):
        a = NestedList([[["v"]]])
        b = a._expand(a.collapse())
        self.assertEqual(b, a.get())

if __name__ == "__main__":
    unittest.main()



import sys; sys.path.append("../")

import unittest
import copy

from delimited.path import TuplePath
from delimited.path import DelimitedStrPath

from delimited import NestedDict


class TestNestedDict(unittest.TestCase):

    # __init__

    def test___init__(self):
        a = NestedDict()
        self.assertEqual(a.data, {})
    
    def test___init____data_arg(self):
        a = NestedDict({"k": {"k": "v"}})
        self.assertEqual(a, {"k": {"k": "v"}})
        
    def test___init____raises_TypeError(self):
        with self.assertRaises(TypeError):
            a = NestedDict("foo")
    
    # __call__
    
    def test___call__(self):
        a = NestedDict({"k": {"k": "v"}})
        a()
        self.assertEqual(a, {})
    
    def test___call____data_arg(self):
        a = NestedDict()
        a({"k": {"k": "v"}})
        self.assertEqual(a, {"k": {"k": "v"}})
    
    # __bool__
    
    def test___bool____return_False(self):
        self.assertFalse(bool(NestedDict()))
    
    def test___bool____return_True(self):
        self.assertTrue(bool(NestedDict({"k": {"k": "v"}})))
    
    # __repr__
    
    def test___repr__(self):
        a = NestedDict({"k": {"k": "v"}})
        self.assertEqual(repr(a), "NestedDict({'k': NestedDict({'k': 'v'})})")
    
    # __eq__
    
    def test___eq____return_True(self):
        a1 = NestedDict({"k": {"k": "v"}})
        a2 = NestedDict({"k": {"k": "v"}})
        self.assertTrue(a1 == a2)
    
    def test___eq____return_False(self):
        a1 = NestedDict()
        a2 = None
        self.assertFalse(a1 == a2)
    
    def test___eq____same_class__return_False(self):
        a1 = NestedDict({"k": {"foo": "bar"}})
        a2 = NestedDict({"k": {"bar": "baz"}})
        self.assertFalse(a1 == a2)
    
    def test___eq____different_class__return_True(self):
        a1 = NestedDict({"k": {"foo": "bar"}})
        a2 = {"k": {"foo": "bar"}}
        self.assertTrue(a1 == a2)
    
        a3 = NestedDict()
        a4 = {}
        self.assertTrue(a3 == a4)
    
    # __ne__
    
    def test___ne____return_True(self):
        a1 = NestedDict({"k": {"foo": "bar"}})
        a2 = NestedDict({"k": {"bar": "baz"}})
        self.assertTrue(a1 != a2)
    
    def test___ne____return_False(self):
        a1 = NestedDict({"k": {"foo": "bar"}})
        a2 = NestedDict({"k": {"foo": "bar"}})
        self.assertFalse(a1 != a2)
    
    # __str__
    
    def test___str__(self):
        a = NestedDict({"foo": "bar"})
        self.assertEqual(str(a), "{'foo': 'bar'}")
    
    # __hash__
    
    def test___hash__(self):
        a1 = hash(NestedDict({"k": {"foo": "bar"}}))
        a2 = hash(NestedDict({"k": {"foo": "bar"}}))
        self.assertTrue(a1 == a2)
    
    # __iter__
    
    def test___iter__(self):
        a = NestedDict({"k1": "v", "k2": "v", "k3": "v"})
        for k, v in a:
            self.assertTrue(k in ["k1", "k2", "k3"])
            self.assertEqual(v, "v")
    
    # __contains__
    
    def test___contains__(self):
        a = NestedDict({"k1": "v", "k2": "v", "k3": "v"})
        self.assertTrue("k1" in a)
    
    # __getitem__
    
    def test___getitem__(self):
        a = NestedDict({"k": {"k": "v"}})
        self.assertEqual(a["k"], {"k": "v"})
    
    # __setitem__
    
    def test___setitem__(self):
        a = NestedDict({"k": {"foo": "bar"}})
        a["k"] = {"bar": "baz"}
        self.assertEqual(a.data, {"k": {"bar": "baz"}})
    
    # __delitem__
    
    def test___delitem__(self):
        a = NestedDict({"k": {"k": "v"}})
        del a["k"]
        self.assertEqual(a.data, {})

    # __len__
    
    def test___len__(self):
        a = NestedDict({"k1": "v", "k2": "v", "k3": "v"})
        self.assertEqual(len(a), 3)
    
    # __copy__
    
    def test___copy__(self):
        a1 = NestedDict({"k": {"foo": "bar"}})
        a2 = copy.copy(a1)
        self.assertIsNot(a1, a2)
        self.assertEqual(a1, a2)
        a1.data["k"]["foo"] = "baz"
        self.assertEqual(a1, a2)
    
    # __deepcopy__
    
    def test___deepcopy__(self):
        a1 = NestedDict({"k": {"foo": "bar"}})
        a2 = copy.deepcopy(a1)
        self.assertIsNot(a1, a2)
        self.assertEqual(a1, a2)
        a1.data["k"]["foo"] = "baz"
        self.assertNotEqual(a1, a2)
    
    # __items__
    
    def test___items__(self):
        a = NestedDict({"k": {"foo": "bar"}})
        for item in a.items():
            self.assertIsInstance(item, tuple)
    
    # __keys__
    
    def test___keys__(self):
        a = NestedDict({"k": {"foo": "bar"}})
        for k in a.keys():
            self.assertEqual(k, "k")
    
    # __values__
    
    def test___values__(self):
        a = NestedDict({"k": {"foo": "bar"}})
        for v in a.values():
            self.assertEqual(v, {"foo": "bar"})
    
    # ref
    
    def test_ref__no_args__returns_all_values(self):
        a = NestedDict({"k": {"foo": "bar"}})
        self.assertEqual(a.ref(), {"k": {"foo": "bar"}})
    
    def test_ref__string_arg__returns_value(self):
        a = NestedDict({"k": {"foo": "bar"}})
        self.assertEqual(a.ref("k"), {"foo": "bar"})
    
    def test_ref__string_arg__raises_KeyError(self):
        a = NestedDict()
        with self.assertRaises(KeyError):
            a.ref("k")
    
    def test_ref__string_arg__raises_TypeError(self):
        a = NestedDict({"k": {"foo": "bar"}})
        with self.assertRaises(TypeError):
            a.ref(("k", "foo", "baz"))
    
    def test_ref__path_arg__returns_value(self):
        a = NestedDict({"k1": {"k2": {"k3": "v"}}})
        self.assertEqual(a.ref(("k1", "k2")), ({"k3": "v"}))
    
    def test_ref__create_True__creates_missing_containers(self):
        a = NestedDict()
        a.ref("k", create=True)
        self.assertEqual(a, {"k": {}})
    
    # get
    
    def test_get__string_arg__returns_value(self):
        a = NestedDict({"k": "v"})
        self.assertEqual(a.get("k"), "v")
    
    def test_get__string_arg_missing_key__returns_default_value(self):
        a = NestedDict()
        self.assertEqual(a.get("k", "foo"), "foo")
    
    def test_get__string_arg__raises_KeyError(self):
        a = NestedDict()
        with self.assertRaises(KeyError):
            a.get("k")
    
    def test_get__string_arg__raises_ValueError(self):
        a = NestedDict({"k1": {"k2": "v"}})
        with self.assertRaises(TypeError):
            a.get(("k1", "k2", "k3"))
    
    def test_get__path_arg__returns_value(self):
        a = NestedDict({"k1": {"k2": {"k3": "v"}}})
        self.assertEqual(a.get(("k1", "k2")), {"k3": "v"})
    
    # has
    
    def test_has__returns_True(self):
        a = NestedDict({"k": "v"})
        self.assertTrue(a.has())
    
    def test_has__returns_False(self):
        a = NestedDict()
        self.assertFalse(a.has())
    
    def test_has__tuple_arg__returns_True(self):
        a = NestedDict({"k1": {"k2": {"k3": "v"}}})
        self.assertTrue(a.has(("k1",)))
    
    def test_has__tuple_arg__returns_False(self):
        a = NestedDict({("k1", "k2", "k3"): "v"})
        self.assertFalse(a.has(("foo",)))
    
    def test_has__path_arg__returns_True(self):
        a = NestedDict({"k1": {"k2": {"k3": "v"}}})
        self.assertTrue(a.has(("k1", "k2", "k3")))
    
    def test_has__path_arg__returns_False(self):
        a = NestedDict({"k1": {"k2": {"k3": "v"}}})
        self.assertFalse(a.has(("k1", "k2", "foo")))
    
    # copy
    
    def test_copy(self):
        a1 = NestedDict({"k1": {"k2": {"k3": "v"}}})
        a2 = a1.copy()
        self.assertEqual(a1, a2)
        a1["k1"] = "foo"
        self.assertNotEqual(a1, a2)
    
    # clone
    
    def test_clone(self):
        a1 = NestedDict({"k1": {"k2": {"k3": "v"}}})
        a2 = a1.clone()
        self.assertEqual(a1, a2)
        a1["k1"] = "foo"
        self.assertEqual(a1, a2)
    
    # set
    
    def test_set__string_arg(self):
        a = NestedDict()
        a.set("k", "v")
        self.assertEqual(a, {"k": "v"})
    
    def test_set__delimited_string_key_param(self):
        a = NestedDict()
        a.set(("k1", "k2", "k3"), "v")
        self.assertEqual(a, {"k1": {"k2": {"k3": "v"}}})
    
    def test_set__raises_TypeError(self):
        a = NestedDict({"k1": {"k2": "v"}})
        with self.assertRaises(TypeError):
            a.set(("k1", "k2", "k3"), "v", create=False)
    
    # push
    
    def test_push__string_arg(self):
        a = NestedDict({"k": []})
        a.push("k", "v")
        self.assertEqual(a, {"k": ["v"]})
    
    def test_push__string_arg__convert_existing_value(self):
        a = NestedDict({"k": "v"})
        a.push("k", "v")
        self.assertEqual(a, {"k": ["v", "v"]})
    
    def test_push__path_arg(self):
        a = NestedDict({"k1": {"k2": {"k3": []}}})
        a.push(("k1", "k2", "k3"), "v")
        self.assertEqual(a, {"k1": {"k2": {"k3": ["v"]}}})
    
    def test_push__create_True__creates_list(self):
        a = NestedDict()
        a.push("k", "v")
        self.assertEqual(a, {"k": ["v"]})
    
    def test_push__create_False__raises_KeyError(self):
        a = NestedDict()
        with self.assertRaises(KeyError):
            a.push("k", "v", create=False)
    
    def test_push__create_False__raises_TypeError(self):
        a = NestedDict({"k": "v"})
        with self.assertRaises(AttributeError):
            a.push("k", "v", create=False)
    
    # pull
    
    def test_pull__string_arg(self):
        a = NestedDict({"k": ["v"]})
        a.pull("k", "v")
        self.assertEqual(a, {"k": []})
    
    def test_pull__cleanup_True__removes_empty_containers(self):
        a = NestedDict({"k": ["v"]})
        a.pull("k", "v", cleanup=True)
        self.assertEqual(a, {})
    
    def test_pull__path_arg(self):
        a = NestedDict({"k1": {"k2": {"k3": ["v"]}}})
        a.pull(("k1", "k2", "k3"), "v")
        self.assertEqual(a, {"k1": {"k2": {"k3": []}}})
    
    def test_pull__string_arg__raises_KeyError(self):
        a = NestedDict()
        with self.assertRaises(KeyError):
            a.pull("k", "v")
    
    def test_pull__string_arg__raises_ValueError(self):
        a = NestedDict({"k": []})
        with self.assertRaises(ValueError):
            a.pull("k", "v")
    
    def test_pull__string_arg__raises_TypeError(self):
        a = NestedDict({"k": "v"})
        with self.assertRaises(AttributeError):
            a.pull("k", "v")
    
    # unset
    
    def test_unset__string_arg(self):
        a = NestedDict({"k": "v"})
        a.unset("k")
        self.assertEqual(a, {})
    
    def test_unset__cleanup_True__removes_empty_containers(self):
        a = NestedDict({"k1": {"k2": {"k3": "v"}}})
        a.unset(("k1", "k2", "k3"), cleanup=True)
        self.assertEqual(a, {"k1": {}})
    
    def test_unset__string_arg__raises_KeyError(self):
        a = NestedDict({"k": "v"})
        with self.assertRaises(KeyError):
            a.unset("foo")
    
    # merge
    
    def test_merge__dict_arg(self):
        a = NestedDict({"k1": "v"})
        b = a.merge({"k2": "v"})
        self.assertEqual(b, {"k1": "v", "k2": "v"})
    
    def test_merge__nested_data__nested_data_arg(self):
        a = NestedDict({"k1": {"k2": "v"}})
        b = a.merge({"k1": {"k3": "v"}})
        self.assertEqual(b, {"k1": {"k2": "v", "k3": "v"}})
    
    def test_merge__NestedDict_arg(self):
        a = NestedDict({"k1": "v"})
        b = a.merge(NestedDict({"k2": "v"}))
        self.assertEqual(b, {"k1": "v", "k2": "v"})
    
    def test_merge__incompatible_types(self):
        a = NestedDict({"k1": "v"})
        b = a.merge("foo")
        self.assertEqual(b, {"k1": "v"})
    
    # update
    
    def test_update__dict_arg(self):
        a = NestedDict({"k1": "v"})
        a.update({"k2": "v"})
        self.assertEqual(a, {
            "k1": "v",
            "k2": "v"
        })
    
    def test_update__NestedDict_arg(self):
        a = NestedDict({"k1": "v"})
        a.update(NestedDict({"k2": "v"}))
        self.assertEqual(a, {
            "k1": "v",
            "k2": "v"
        })
    
    # collapse
    
    def test_collapse(self):
        a = NestedDict({"k1": {"k2": {"k3": "v"}}})
        b = a.collapse()
        self.assertEqual(b, {("k1", "k2", "k3"): "v"})
    
    def test_collapse__function_arg(self):
        a = NestedDict({"k1": {"k2": {"$foo": {"k3": {"k4": "v"}}}}})
    
        def detect_operator(key, *args):
            return key[0] == "$"
    
        b = a.collapse(func=detect_operator)
        self.assertEqual(b, {("k1", "k2"): {"$foo": {("k3", "k4"): "v"}}})
        
    def test_collapse__function_arg__stop_at_root(self):
        a = NestedDict({"k1": {"k2": {"$foo": {"k3": {"k4": "v"}}}}})
    
        def detect_operator(key, *args):
            return key == "k1"
    
        b = a.collapse(func=detect_operator)
        self.assertEqual(b, {"k1": {("k2", "$foo", "k3", "k4"): "v"}})
        
    # expand
    
    def test_expand(self):
        a = NestedDict({"k1": {"k2": {"k3": "v"}}})
        b = NestedDict().expand(a.collapse())
        self.assertEqual(b, a)
        
    def test__expand(self):
        a = NestedDict({"k1": {"k2": {"k3": "v"}}})
        b = a._expand(a.collapse())
        self.assertEqual(b, a.get())
        

if __name__ == "__main__":
    unittest.main()



"""
delimited
~~~~~~~~~
"""

from delimited.type import NestedType
from delimited.path import TuplePath
from delimited.path import DelimitedStrPath


__version__ = "0.0.10"
__title__ = "delimited"
__summary__ = "Defines types that allow for accessing and modifying nested data"
__url__ = "https://github.com/chrisantonellis/delimited"
__author__ = "Christopher Antonellis"
__email__ = "christopher.antonellis@gmail.com"
__license__ = "MIT License"


NestedDict, NestedList = NestedType("nested", TuplePath)
DelimitedDict, DelimitedList = NestedType("delimited", DelimitedStrPath)


import sys; sys.path.append("../")

import unittest
import copy

from delimited.path import TuplePath
from delimited.path import DelimitedStrPath

from delimited.sequence import ListIndex


class TestTuplePath(unittest.TestCase):

    # _encode

    def test___encode(self):
        a = TuplePath()
        self.assertEqual(a._encode(["k1", "k2", "k3"]), ("k1", "k2", "k3"))

    # _decode

    def test___decode(self):
        a = TuplePath()
        self.assertEqual(a._decode(("k1", "k2", "k3")), ["k1", "k2", "k3"])

    # __init__

    def test___init___no_params(self):
        a = TuplePath()
        self.assertEqual(a, ())

    def test___init___tuple_key_notation_param(self):
        a = TuplePath(("k1", "k2", "k3"))
        self.assertEqual(a, ("k1", "k2", "k3"))

    # __call__

    def test___call___no_params(self):
        a = TuplePath()
        a()
        self.assertEqual(a, ())

    def test___call___no_params__existing_segments(self):
        a = TuplePath()
        a.segments = ["foo", "bar", "baz"]
        a()
        self.assertEqual(a, ())

    def test___call___tuple_key_notation_param__existing_segments(self):
        a = TuplePath()
        a.segments = ["foo", "bar", "baz"]
        a(("k1", "k2", "k3"))
        self.assertEqual(a, ("k1", "k2", "k3"))

    # __add__

    def test___add__(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = "k4"
        c = a + b
        self.assertIsInstance(c, TuplePath)
        self.assertEqual(c, ("k1", "k2", "k3", "k4"))

    # __iadd__

    def test___iadd____tuple_param(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = "k4"
        a += b
        self.assertEqual(a, ("k1", "k2", "k3", "k4"))

    # __eq__

    def test___eq___same_class_param__returns_True(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = TuplePath(("k1", "k2", "k3"))
        self.assertTrue(a == b)

    def test___eq___same_class_param__returns_False(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = TuplePath(("foo", "bar", "baz"))
        self.assertFalse(a == b)

    def test___eq___tuple_param__returns_True(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = ("k1", "k2", "k3")
        self.assertTrue(a == b)

    def test___eq___tuple_param__returns_False(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = ("foo", "bar", "baz")
        self.assertFalse(a == b)

    def test___eq___incompatible_type__returns_False(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = True
        self.assertFalse(a == b)

    # __ne__

    def test___ne___same_class_param__returns_False(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = TuplePath(("k1", "k2", "k3"))
        self.assertFalse(a != b)

    def test___ne___same_class_param__returns_True(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = TuplePath(("foo", "bar", "baz"))
        self.assertTrue(a != b)

    def test___ne___tuple_param__returns_False(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = ("k1", "k2", "k3")
        self.assertFalse(a != b)

    def test___ne___tuple_param__returns_True(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = ("foo", "bar", "baz")
        self.assertTrue(a != b)

    def test___ne___incompatible_type__returns_True(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = Exception
        self.assertTrue(a != b)

    # __hash__

    def test___hash__(self):
        a = hash(TuplePath(("k1", "k2", "k3")))
        b = hash(TuplePath(("k1", "k2", "k3")))
        self.assertEqual(a, b)

    # __len__

    def test___len__(self):
        a = TuplePath()
        self.assertEqual(len(a), 0)

        b = TuplePath(("k1", "k2", "k3"))
        self.assertEqual(len(b), 3)

    # __str__

    def test___str__(self):
        a = TuplePath()
        self.assertEqual(str(a), "()")
        b = TuplePath(("k1", "k2", "k3"))
        self.assertEqual(str(b), "('k1', 'k2', 'k3')")

    # __repr__

    def test___repr__(self):
        a = TuplePath(("k1", "k2", "k3"))
        self.assertEqual(repr(a), "TuplePath(('k1', 'k2', 'k3'))")

    # __bool__

    def test___bool___returns_False(self):
        a = TuplePath()
        self.assertFalse(a)

    def test___bool___returns_True(self):
        a = TuplePath(("k1", "k2", "k3"))
        self.assertTrue(a)

    # __iter__

    def test___iter__(self):
        a = TuplePath(("k1", "k2", "k3"))
        self.assertEqual(list(iter(a)), ["k1", "k2", "k3"])

    # __reversed__

    def test___reversed__(self):
        a = TuplePath(("k1", "k2", "k3"))
        self.assertEqual(list(reversed(a)), ["k3", "k2", "k1"])

    # __contains__

    def test___contains____returns_True(self):
        a = TuplePath(("k1", "k2", "k3"))
        self.assertTrue("k2" in a)

    def test___contains____returns_False(self):
        a = TuplePath(("k1", "k2", "k3"))
        self.assertFalse("foo" in a)

    # __getitem__

    def test___getitem__int_index(self):
        a = TuplePath(("k3", "k2", "k1"))
        self.assertEqual(a[1], "k2")

    def test___getitem__slice_index(self):
        a = TuplePath(("k3", "k2", "k1"))
        self.assertEqual(a[:-1], ("k3", "k2"))

    # __setitem__

    def test___setitem__int_index(self):
        a = TuplePath(("k1", "k2", "k3"))
        a[1] = "foo"
        self.assertEqual(a, ("k1", "foo", "k3"))

    def test___setitem__slice_index(self):
        a = TuplePath(("k1", "k2", "k3"))
        a[:-1] = ("foo", "bar")
        self.assertEqual(a, ("foo", "bar", "k3"))

    # __delitem__

    def test___delitem__int_index(self):
        a = TuplePath(("k1", "k2", "k3"))
        del a[1]
        self.assertEqual(a, ("k1", "k3"))

    def test___delitem__slice_index(self):
        a = TuplePath(("k1", "k2", "k3"))
        del a[:-1]
        self.assertEqual(a, ("k3",))

    # append

    def test_append__str_param(self):
        a = TuplePath()
        self.assertEqual(a, ())
        a.append("k1")
        self.assertEqual(a, ("k1",))
        a.append("k2")
        self.assertEqual(a, ("k1", "k2"))

    # extend

    def test_extend__same_class(self):
        a = TuplePath(("k1",))
        b = TuplePath(("k2",))
        a.extend(b)
        self.assertEqual(a, ("k1", "k2"))

    def test_extend__external_representation(self):
        a = TuplePath(("k1",))
        b = ("k2",)
        a.extend(b)
        self.assertEqual(a, ("k1", "k2"))

    # insert

    def test_insert(self):
        a = TuplePath(("k2",))
        a.insert(1, "k3")
        self.assertEqual(a, ("k2", "k3"))
        a.insert(0, "k1")
        self.assertEqual(a, ("k1", "k2", "k3"))

    # remove

    def test_remove(self):
        a = TuplePath(("k1", "k2", "k3"))
        a.remove("k2")
        self.assertEqual(a, ("k1", "k3"))

    # pop

    def test_pop(self):
        a = TuplePath(("k1", "k2", "k3"))
        self.assertEqual(a.pop(), "k3")
        self.assertEqual(a, ("k1", "k2"))
        self.assertEqual(a.pop(0), "k1")
        self.assertEqual(a, ("k2",))

    # clear

    def test_clear(self):
        a = TuplePath(("k1", "k2", "k3"))
        a.clear()
        self.assertEqual(a, ())

    # index

    def test_index(self):
        a = TuplePath(("k1", "k2", "k3"))
        self.assertEqual(a.index("k2"), 1)

    # count

    def test_count(self):
        a = TuplePath(("k1", "k1", "k2"))
        self.assertEqual(a.count("k1"), 2)

    # reverse

    def test_reverse(self):
        a = TuplePath(("k1", "k2", "k3"))
        a.reverse()
        self.assertEqual(a, ("k3", "k2", "k1"))

    # copy

    def test_copy(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = a.copy()
        self.assertIsNot(a, b)
        self.assertEqual(a, b)
        a[1] = "foo"
        self.assertEqual(a, ("k1", "foo", "k3"))
        self.assertNotEqual(a, b)
        
    def test_copy__index_set(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = a.copy(slice(1, len(a)))
        self.assertEqual(b, ("k2", "k3"))
        
    # clone
    
    def test_clone(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = a.clone()
        self.assertIsNot(a, b)
        self.assertEqual(a, b)
        a[1] = "foo"
        self.assertEqual(a, ("k1", "foo", "k3"))
        self.assertEqual(b, ("k1", "foo", "k3"))
        self.assertEqual(a, b)

    # encode

    def test_encode(self):
        a = TuplePath(("k1", "k2", "k3"))
        b = a.encode()
        self.assertIsInstance(b, tuple)
        self.assertEqual(b, ("k1", "k2", "k3"))


class TestDelimitedStrPath(unittest.TestCase):

    # _encode

    def test__encode(self):
        a = DelimitedStrPath()
        self.assertEqual(a._encode(["k1", "k2", "k3"]), "k1.k2.k3")
        
    def test__encode__non_list_arg(self):
        a = DelimitedStrPath()
        self.assertEqual(a._encode("k"), "k")
        
    def test__encode__list_index(self):
        a = DelimitedStrPath()
        self.assertEqual(a._encode(["k1", ListIndex(0), "k2"]), "k1.::0.k2")

    # _decode

    def test__decode(self):
        a = DelimitedStrPath()
        self.assertEqual(a._decode("k1.k2.k3"), ["k1", "k2", "k3"])
        
    def test__decode__list_index(self):
        a = DelimitedStrPath()
        self.assertEqual(a._decode("k1.::0.k3"), ["k1", ListIndex(0), "k3"])
        
    # __repr__
    
    def test____repr__(self):
        a = DelimitedStrPath()
        self.assertEqual(repr(a), "DelimitedStrPath('')")
        
        a("k1.k2.k3")
        self.assertEqual(repr(a), "DelimitedStrPath('k1.k2.k3')")


if __name__ == "__main__":
    unittest.main()



"""
delimited.type
~~~~~~~~~~~~~~~~~

This module defines...
"""

from delimited.container import NestedContainer
from delimited.mapping import NestedMapping
from delimited.sequence import NestedSequence
from delimited.sequence import ListIndex


class NestedType(object):

    def __new__(cls, name, path):

        mapping = type(
            f"{name.title()}Dict",
            (NestedMapping, NestedContainer),
            {
                "container": dict,
                "path": path
            }
        )

        sequence = type(
            f"{name.title()}List",
            (NestedSequence, NestedContainer),
            {
                "container": list,
                "path": path,
                "sequenceindex": ListIndex,
            }
        )
        
        mapping.mapping = sequence.mapping = mapping
        mapping.sequence = sequence.sequence = sequence
        
        return mapping, sequence
